#!/usr/bin/env python
# Copyright (c) 2015 Mirantis, Inc
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import functools
import requests
import socket

from six.moves import http_client
from six.moves.urllib import error

from yretry import decorators
from yretry import defaults


RETRY_SOCKET_EXCEPTIONS = (socket.error, socket.timeout)
RETRY_HTTPLIB_EXCEPTIONS = (http_client.BadStatusLine,
                            http_client.NotConnected,
                            http_client.CannotSendRequest,
                            http_client.CannotSendHeader)
RETRY_URLLIB_EXCEPTIONS = (error.HTTPError, )
RETRY_REQUESTS_EXCEPTIONS = (requests.ConnectionError, requests.Timeout)
RETRY_HTTP_CODES = (http_client.REQUEST_TIMEOUT,
                    http_client.INTERNAL_SERVER_ERROR,
                    http_client.BAD_GATEWAY,
                    http_client.SERVICE_UNAVAILABLE,
                    http_client.GATEWAY_TIMEOUT,)


def is_retriable_requests_httperror(error):
    """Returns true error is retriable requests.exceptions.HTTPError."""
    return (isinstance(error, requests.exceptions.HTTPError) and
            error.response.status_code in RETRY_HTTP_CODES)


def is_network_failure(error):
    """Returns True when error is a network failure."""
    return ((isinstance(error, RETRY_URLLIB_EXCEPTIONS)
             and error.code in RETRY_HTTP_CODES) or
            isinstance(error, RETRY_HTTPLIB_EXCEPTIONS) or
            isinstance(error, RETRY_SOCKET_EXCEPTIONS) or
            isinstance(error, RETRY_REQUESTS_EXCEPTIONS) or
            is_retriable_requests_httperror(error))


def retry(attempts_number=None,
          delay=None,
          step=0,
          retry_on=None,
          retry_except=None,
          logger=None):

    def decorator(func):

        @functools.wraps(func)
        def wrapper(*args, **kwargs):

            if attempts_number is not None:
                retry_attempts = attempts_number
            else:
                retry_attempts = defaults.HTTP_RETRY_ATTEMPTS

            if delay is not None:
                retry_delay = delay
            else:
                retry_delay = defaults.HTTP_RETRY_DELAY

            if retry_on is None:
                retry_func = is_network_failure
            else:
                retry_func = retry_on

            failover_func = decorators.retry(retry_attempts,
                                             retry_delay,
                                             step,
                                             retry_func,
                                             retry_except,
                                             logger)(func)
            return failover_func(*args, **kwargs)

        return wrapper

    return decorator



#!/usr/bin/env python
# Copyright (c) 2015 Mirantis, Inc
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import socket

import mock
from six.moves import http_client
from six.moves.urllib import error

from yretry import network
from yretry.tests import base


RETRY_ATTEMPTS = 3


class FakeClass(mock.Mock):

    @network.retry(RETRY_ATTEMPTS, delay=0)
    def retry_method_works_incorrect(self):
        self.retry_count()
        raise socket.timeout()

    @network.retry(RETRY_ATTEMPTS, delay=0)
    def retry_method_works_incorrect_http_exception(self):
        self.retry_count()
        raise error.HTTPError(
            "FakeUrl", http_client.REQUEST_TIMEOUT,
            "FakeMessage", None, None)

    @network.retry(RETRY_ATTEMPTS, delay=0)
    def retry_method_works_correct(self):
        self.retry_count()


class RetryTestCase(base.TestCase):

    def setUp(self):
        self._target = FakeClass()

    def test_must_retry_three_time_and_raise_exception(self):
        self.assertRaises(
            socket.timeout,
            self._target.retry_method_works_incorrect)
        self.assertEqual(self._target.retry_count.call_count, 3)

    def test_must_retry_three_time_and_raise_http_exception(self):
        self.assertRaises(
            error.HTTPError,
            self._target.retry_method_works_incorrect_http_exception)
        self.assertEqual(self._target.retry_count.call_count, 3)

    def test_must_retry_one_time_and_return_correct_result(self):
        self.assertIsNone(self._target.retry_method_works_correct())
        self.assertEqual(self._target.retry_count.call_count, 1)



#!/usr/bin/env python
# Copyright (c) 2015 Mirantis, Inc
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import mock

from yretry import decorators
from yretry.tests import base


RETRY_ATTEMPTS = 3

START_DELAY = 0

STEP = 1


class SuperPuperException(Exception):
    pass


class DontRetryException(Exception):
    pass


class FakeClass(mock.Mock):

    @decorators.retry(RETRY_ATTEMPTS, delay=0)
    def retry_method_works_incorrect(self):
        self.retry_count()
        raise SuperPuperException()

    @decorators.retry(RETRY_ATTEMPTS, delay=0,
                      retry_on=(SuperPuperException,))
    def retry_method_works_incorrect_2(self):
        self.retry_count()
        raise DontRetryException()

    @decorators.retry(RETRY_ATTEMPTS, delay=0)
    def retry_method_works_correct(self):
        self.retry_count()

    @decorators.retry(RETRY_ATTEMPTS, delay=START_DELAY,
                      step=STEP)
    def retry_with_custom_step(self):
        self.retry_count()
        raise SuperPuperException()

    @decorators.retry(RETRY_ATTEMPTS, delay=0,
                      retry_except=(SuperPuperException,))
    def retry_on_except(self):
        self.retry_count()
        raise SuperPuperException()


class RetryTestCase(base.TestCase):

    def setUp(self):
        self._target = FakeClass()

    def test_must_retry_three_time_and_raise_exception(self):
        self.assertRaises(
            SuperPuperException,
            self._target.retry_method_works_incorrect)
        self.assertEqual(self._target.retry_count.call_count,
                         RETRY_ATTEMPTS)

    def test_must_retry_one_time_and_return_correct_result(self):
        self.assertIsNone(self._target.retry_method_works_correct())
        self.assertEqual(self._target.retry_count.call_count, 1)

    def test_dont_retry_on_unexpected_error(self):
        """Don't retry if exception isn't SuperPuperException."""
        self.assertRaises(
            DontRetryException,
            self._target.retry_method_works_incorrect_2)
        self.assertEqual(self._target.retry_count.call_count, 1)

    @mock.patch('time.sleep')
    def test_step_is_works(self, sleep):
        """retry delay increases to step value."""
        self.assertRaises(
            SuperPuperException,
            self._target.retry_with_custom_step)
        sleep.assert_has_calls([
            mock.call(START_DELAY),
            mock.call(START_DELAY + STEP)
        ])

    def test_retry_works_with_function_without_parameters(self):
        @decorators.retry(RETRY_ATTEMPTS, delay=0)
        def function_without_parameters():
            return "OK"

        self.assertEqual(function_without_parameters(), "OK")

    def test_dont_retry_on_expected_error(self):
        """Don't retry if exception is SuperPuperException."""
        self.assertRaises(
            SuperPuperException,
            self._target.retry_on_except,
            )
        self.assertEqual(self._target.retry_count.call_count, 1)


class LoggerTestCase(base.TestCase):

    def setUp(self):
        super(LoggerTestCase, self).setUp()

        self.logger = mock.MagicMock()

    @mock.patch('time.sleep')
    def test_works_without_logger(self, sleep):
        class Counter(object):

            def __init__(self):
                self._value = 0

            @property
            def value(self):
                return self._value

            def inc(self):
                self._value += 1

        @decorators.retry(RETRY_ATTEMPTS,
                          retry_on=(SuperPuperException,))
        def bad_function(counter):
            counter.inc()
            raise SuperPuperException()

        counter = Counter()

        self.assertRaises(SuperPuperException,
                          bad_function,
                          counter)

        self.assertEqual(counter.value, 3)

    @mock.patch('time.sleep')
    def test_predefined_logger_is_used(self, sleep):
        """decorators.retry uses logger to write debug information

        Decorator writes warning messages about retries.
        """
        @decorators.retry(RETRY_ATTEMPTS,
                          logger=self.logger,
                          retry_on=(SuperPuperException,))
        def bad_function():
            raise SuperPuperException()

        self.assertRaises(SuperPuperException,
                          bad_function)

        self.assertEqual(self.logger.warning.call_count,
                         RETRY_ATTEMPTS - 1)

    @mock.patch('time.sleep')
    def test_object_logger_is_used(self, sleep):
        """decorators.retry uses object's logger

        """

        class TestClass(object):

            def __init__(self):
                self.logger = mock.MagicMock()

            def get_logger(self):
                return self.logger

            @decorators.retry(RETRY_ATTEMPTS,
                              retry_on=(SuperPuperException,))
            def reliable_method(self):
                raise SuperPuperException()

        obj = TestClass()

        self.assertRaises(SuperPuperException,
                          obj.reliable_method)

        self.assertEqual(obj.logger.warning.call_count,
                         RETRY_ATTEMPTS - 1)

    @mock.patch('time.sleep')
    def test_object_logger_is_used_predefined_is_ignored(self, sleep):
        """decorators.retry ignores predefined logger

        In case when object logger is defined.
        """

        class TestClass(object):

            def __init__(self):
                self.logger = mock.MagicMock()

            def get_logger(self):
                return self.logger

            @decorators.retry(RETRY_ATTEMPTS,
                              logger=self.logger,
                              retry_on=(SuperPuperException,))
            def reliable_method(self):
                raise SuperPuperException()

        obj = TestClass()

        self.assertRaises(SuperPuperException,
                          obj.reliable_method)

        self.assertEqual(self.logger.warning.call_count,
                         0)

        self.assertEqual(obj.logger.warning.call_count,
                         RETRY_ATTEMPTS - 1)

    @mock.patch('time.sleep')
    def test_works_if_get_logger_isnt_defined(self, sleep):
        """decorators works if get_logger isn't defined."""

        class TestClass(object):

            def __init__(self):
                pass

            @decorators.retry(RETRY_ATTEMPTS,
                              retry_on=(SuperPuperException,))
            def reliable_method(self):
                raise SuperPuperException()

        obj = TestClass()

        self.assertRaises(SuperPuperException,
                          obj.reliable_method)



#!/usr/bin/env python
# Copyright (c) 2014 Sergey Bunatyan <sbunatyan@mirantis.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import abc
import functools
import time
import types

import six


@six.add_metaclass(abc.ABCMeta)
class CatchStrategy(object):

    @abc.abstractmethod
    def need_to_retry(self, exc):
        pass


class CatchFunctionStrategy(CatchStrategy):

    def __init__(self, to_retry):
        super(CatchFunctionStrategy, self).__init__()
        self._to_retry = to_retry

    def need_to_retry(self, exc):
        return self._to_retry(exc)


class CatchExceptionStrategy(CatchFunctionStrategy):

    def __init__(self, exceptions_to_retry, exceptions_to_ignore=None):
        if exceptions_to_ignore:
            def strategy(exc):
                return (isinstance(exc, exceptions_to_retry)
                        and not isinstance(exc,
                                           exceptions_to_ignore))
        else:
            def strategy(exc):
                return isinstance(exc, exceptions_to_retry)
        super(CatchExceptionStrategy, self).__init__(strategy)


def retry(attempts_number,
          delay=0,
          step=0,
          retry_on=Exception,
          retry_except=None,
          logger=None):
    """Reties function several times

    @param attempts_number: number of function calls (first call + retries)
    @param timeout: timeout before first retry
    @param step: increment value of timeout on each retry
    @param retry_on: exception that should be handled or function that checks
                     if retry should be executed (default: Exception)
    @param retry_except: exception that should not be handled if exception
                         was specified into `retry_on` (default: None)
    @param logger: logger to write warnings

    @return: the result of decorated function
    """

    def decorator(func):

        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            current_logger = logger

            attempts = 1
            retry_delay = delay

            try:
                if isinstance(args[0], object):
                    current_logger = args[0].get_logger()
            except (AttributeError, IndexError):
                pass

            if isinstance(retry_on, (types.FunctionType,
                                     types.MethodType,)):
                catch_strategy = CatchFunctionStrategy(retry_on)
            else:
                # TODO(g.melikov): this retry() should be splitted into
                #  multiple functions (like `retry_on_func` and
                #  `retry_on_exceptions`); current interface is leaky.
                #  Also `retry_except` parameter takes effect only in this
                #  execution branch and is ignored in other.
                catch_strategy = CatchExceptionStrategy(retry_on, retry_except)

            while attempts <= attempts_number:
                try:
                    return func(*args, **kwargs)
                except Exception as e:
                    if catch_strategy.need_to_retry(e):
                        if attempts >= attempts_number:
                            raise
                        elif current_logger:
                            current_logger.warning(
                                "Retry: Call to %(fn)s failed due to "
                                "%(exc_class)s: %(exc)s, retry "
                                "attempt #%(retry_no)s/"
                                "%(retry_count)s after %(delay)ss",
                                dict(fn=func.__name__,
                                     exc=str(e),
                                     retry_no=attempts,
                                     exc_class=e.__class__.__name__,
                                     retry_count=attempts_number - 1,
                                     delay=retry_delay))
                        time.sleep(retry_delay)
                        attempts += 1
                        retry_delay += step
                    else:
                        raise
        return wrapper
    return decorator
